use std::{fmt, sync::Arc, time::Duration};

use error_stack::{Context, IntoReport, Report, ResultExt};
use pyo3::prelude::*;
use serde::{Deserialize, Serialize};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    net::{
        tcp::{OwnedReadHalf, OwnedWriteHalf},
        TcpListener, TcpStream,
    },
    sync::{mpsc, oneshot, watch, Semaphore},
};
use tracing::{error, info, instrument};

#[derive(Debug)]
struct TestError;

impl fmt::Display for TestError {
    fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt.write_str("Test error")
    }
}

impl Context for TestError {}

struct Listener {
    limit_connections: Arc<Semaphore>,
}

impl Listener {
    async fn run(
        &self,
        request: watch::Receiver<Request>,
        response: mpsc::Sender<Result<Response, Report<TestError>>>,
    ) {
        // Bind the listener to the address
        let listener = TcpListener::bind("127.0.0.1:6379").await.unwrap();
        info!("accepting inbound connections");

        loop {
            // Wait for a permit to become available
            //
            // `acquire_owned` returns a permit that is bound to the semaphore.
            // When the permit value is dropped, it is automatically returned
            // to the semaphore.
            let permit = self
                .limit_connections
                .clone()
                .acquire_owned()
                .await
                .unwrap();

            // The second item contains the IP and port of the new connection.
            let (socket, _) = listener.accept().await.unwrap();
            info!("Incoming connection");
            let handler = Handler::new(socket, request.clone(), response.clone());

            tokio::spawn(async move {
                handler.run().await;

                // free semaphore if connection terminated
                info!("Allowing new connections again");
                drop(permit);
            });
        }
    }
}

struct Handler {
    stream_in: OwnedReadHalf,
    stream_out: OwnedWriteHalf,
    request: watch::Receiver<Request>,
    response: mpsc::Sender<Result<Response, Report<TestError>>>,
}

impl Handler {
    pub fn new(
        socket: TcpStream,
        request: watch::Receiver<Request>,
        response: mpsc::Sender<Result<Response, Report<TestError>>>,
    ) -> Self {
        let (stream_in, stream_out) = socket.into_split();

        Self {
            stream_in,
            stream_out,
            request,
            response,
        }
    }

    #[instrument(skip(self))]
    async fn run(self) {
        let (tx1, rx1) = oneshot::channel::<()>();
        let (tx2, rx2) = oneshot::channel::<()>();

        Self::read(self.stream_in, self.response, tx1).await;
        Self::write(self.stream_out, self.request, tx2).await;

        tokio::select! {
            _ = rx1 => {
                println!("read() completed first")
            }
            _ = rx2 => {
                println!("write() completed first")
            }
        };
    }

    async fn read(
        mut stream: OwnedReadHalf,
        tx: mpsc::Sender<Result<Response, Report<TestError>>>,
        exit: oneshot::Sender<()>,
    ) {
        tokio::spawn(async move {
            loop {
                // check if connection is still active
                if exit.is_closed() {
                    break;
                }

                // wait for incoming messages from connection
                if let Ok(size) = stream.read_u32().await {
                    let mut buf = vec![0; size as usize];
                    if let Err(e) = stream.read_exact(&mut buf).await {
                        error!("Error on reading from client: {e:?}");
                        let _ = exit.send(());
                        break;
                    };

                    let msg = String::from_utf8(buf)
                        .into_report()
                        .change_context(TestError)
                        .and_then(|s| {
                            serde_json::from_str::<Response>(&s)
                                .into_report()
                                .change_context(TestError)
                        });

                    let _ = tx.send(msg).await;
                } else {
                    info!("Connection closed by client");
                    let _ = exit.send(());
                    break;
                }
            }
        });
    }
    async fn write(
        mut stream: OwnedWriteHalf,
        mut rx: watch::Receiver<Request>,
        exit: oneshot::Sender<()>,
    ) {
        tokio::spawn(async move {
            loop {
                // check if connection is still active
                if exit.is_closed() {
                    break;
                }
                // wait for incoming requests
                if rx.changed().await.is_ok() {
                    let msg = serde_json::to_string(&*rx.borrow()).unwrap();
                    // send message
                    if let Err(e) = stream.write_all(msg.as_bytes()).await {
                        // close connection if stream is closed
                        error!("Error on writing to stream: {e:?}");
                        let _ = exit.send(());
                        break;
                    }
                }
            }
        });
    }
}

struct Server {
    request: watch::Sender<Request>,
    response: mpsc::Receiver<Result<Response, Report<TestError>>>,
}

impl Server {
    pub async fn new() -> Self {
        let (tx_request, rx_request) = watch::channel(Request::default());
        let (tx_response, rx_response) = mpsc::channel(10);

        tokio::spawn(async move {
            let server = Listener {
                limit_connections: Arc::new(Semaphore::new(1)),
            };

            server.run(rx_request, tx_response).await;
        });

        Self {
            request: tx_request,
            response: rx_response,
        }
    }
}

#[tokio::main]
async fn main() {
    println!("Hello, world!");
    tracing_subscriber::fmt().init();
    let mut server = Server::new().await;

    println!("Server started");
    let mut i = 0;

    let code = include_str!("../../client.py");

    std::thread::spawn(move || loop {
        if let Err(e) = Python::with_gil(|py| -> PyResult<()> {
            PyModule::from_code(py, code, "client", "client")?;
            Ok(())
        }) {
            error!("Error during execution of Python client: {e}");
        };
        std::thread::sleep(Duration::from_secs(5));
        info!("Restarting Python client")
    });

    loop {
        let _ = server.request.send(Request {
            test: format!("msg {i}"),
        });

        tokio::time::sleep(Duration::from_secs(5)).await;
        i += 1;

        if let Ok(msg) = server.response.try_recv() {
            println!("{msg:?}");
        }
    }
}

#[derive(Debug, Default, Serialize)]
struct Request {
    test: String,
}

#[derive(Debug, Deserialize)]
struct Response {
    test: String,
}
